/*
 0 3
 1 4
 2 5
 6 7
*/

use std::u32;
use std::convert::TryFrom;


#[derive(Clone, Copy)]
#[repr(u8)]
pub enum EncodeOrientation {
    // reverse - points in line ordered backwards (direction)
    // Backwards - major lines ordered backwards (wrap)
    Rows                     = 0b00000000,
    RowsBackwards            = 0b00000001,
    RowsReverse              = 0b00000010,
    RowsBackwardsReverse     = 0b00000011,
    Columns                  = 0b00000100,
    ColumnsBackwards         = 0b00000101,
    ColumnsReverse           = 0b00000110,
    ColumnsBackwardsReverse  = 0b00000111,
}

impl TryFrom<u8> for EncodeOrientation {
    
    type Error = ();

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(EncodeOrientation::Rows),
            1 => Ok(EncodeOrientation::RowsBackwards),
            2 => Ok(EncodeOrientation::RowsReverse),
            3 => Ok(EncodeOrientation::RowsBackwardsReverse),
            4 => Ok(EncodeOrientation::Columns),
            5 => Ok(EncodeOrientation::ColumnsBackwards),
            6 => Ok(EncodeOrientation::ColumnsReverse),
            7 => Ok(EncodeOrientation::ColumnsBackwardsReverse),
            _ => panic!("Unknown encode orientation provided")
        }
    }
}


const MAPPINGS: [[u8; 8]; 8] = [
    [0, 3, 1, 4, 2, 5, 6, 7],
    [6, 7, 2, 5, 1, 4, 0, 3],
    [3, 0, 4, 1, 5, 2, 7, 6],
    [7, 6, 5, 2, 4, 1, 3, 0],
    [0, 1, 2, 6, 3, 4, 5, 7],
    [3, 4, 5, 7, 0, 1, 2, 6],
    [6, 2, 1, 0, 7, 5, 4, 3],
    [7, 5, 4, 3, 6, 2, 1, 0]
];

pub fn index_to_braillie_char(_index: u8, mode: EncodeOrientation) -> char {
    let mapping = MAPPINGS[mode as usize];
    let index: u8 =
        ((_index >> 7) & 1) << mapping[0]
      | ((_index >> 6) & 1) << mapping[1]
      | ((_index >> 5) & 1) << mapping[2]
      | ((_index >> 4) & 1) << mapping[3]
      | ((_index >> 3) & 1) << mapping[4]
      | ((_index >> 2) & 1) << mapping[5]
      | ((_index >> 1) & 1) << mapping[6]
      | ((_index >> 0) & 1) << mapping[7];
    return char::from_u32(0x2800 + index as u32).unwrap();
}

pub fn to_binary_braillie(input: &[u8], length: usize, mode: EncodeOrientation) -> String {
    let mut rvalue = String::with_capacity(length);
    for i in 0..length {
        rvalue.push(index_to_braillie_char(input[i], mode));
    }
    return rvalue;
}

pub fn to_binary_braillie_str(input: impl AsRef<str>, mode: EncodeOrientation) -> String {
    return to_binary_braillie(input.as_ref().as_bytes(), input.as_ref().len(), mode);
}

pub fn binary_braillie_char_to_index(input: char, mode: EncodeOrientation) -> Result<u8, ()>{
    if !(0x2800 <= input as u32 && input as u32 <= 0x28ff) { return Err(()) }
    let mapping = MAPPINGS[mode as usize];
    let _index: u8 = (input as u32 - 0x2800) as u8;
    return Ok(
        ((_index >> mapping[7]) & 1) << 0
      | ((_index >> mapping[6]) & 1) << 1
      | ((_index >> mapping[5]) & 1) << 2
      | ((_index >> mapping[4]) & 1) << 3
      | ((_index >> mapping[3]) & 1) << 4
      | ((_index >> mapping[2]) & 1) << 5
      | ((_index >> mapping[1]) & 1) << 6
      | ((_index >> mapping[0]) & 1) << 7
    );
}

pub fn binary_braillie_to_bytes(
    input: impl AsRef<str>,
    mode: EncodeOrientation
) -> Result<Vec<u8>, ()> {
    let mut rvalue: Vec<u8> = Vec::with_capacity(input.as_ref().chars().count());
    for char in input.as_ref().chars() {
        rvalue.push(match binary_braillie_char_to_index(char, mode) {
            Ok(e) => e,
            Err(e) => return Err(e),
        })
    }
    return Ok(rvalue);
}

pub fn binary_braillie_to_str(input: impl AsRef<str>, mode: EncodeOrientation) -> Result<String, ()> {
    let bytes = match binary_braillie_to_bytes(input, mode) {
        Ok(e) => e,
        Err(e) => return Err(e)
    };
    return match String::from_utf8(bytes) {
        Ok(e) => Ok(e),
        Err(_e) => Err(())
    }
}

#[cfg(test)]
mod test_lib_to_braillie {
    use crate::{index_to_braillie_char, to_binary_braillie_str, EncodeOrientation};

    #[test]
    fn check_characters_in_range() {
        for i in 0..0xff {
            let char = index_to_braillie_char(i, EncodeOrientation::Rows);
            assert!(0x2800 <= char as u32 && char as u32 <= 0x28ff);
        }
    }

    #[test]
    fn check_individual_bits() {
        assert_eq!(
            index_to_braillie_char(0b10000000, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b00000001).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b01000000, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b00001000).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b00100000, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b00000010).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b00010000, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b00010000).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b00001000, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b00000100).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b00000100, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b00100000).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b00000010, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b01000000).unwrap()
        );
        assert_eq!(
            index_to_braillie_char(0b00000001, EncodeOrientation::Rows),
            char::from_u32(0x2800 + 0b10000000).unwrap()
        );
    }

    #[test]
    fn check_some_strings() {
        assert_eq!(
            to_binary_braillie_str("Hi", EncodeOrientation::Rows),
            "⠌⢎"
        );
        assert_eq!(
            to_binary_braillie_str("Hello, world!", EncodeOrientation::Rows),
            "⠌⢪⠮⠮⣮⠦⠂⣺⣮⡚⠮⠪⢂"
        );
        assert_eq!(
            to_binary_braillie_str("ąęćó", EncodeOrientation::Rows),
            "⠩⢡⠩⢕⠩⣡⣉⣓"
        );
    }
}

#[cfg(test)]
mod test_lib_from_braillie {
    use crate::{binary_braillie_char_to_index, EncodeOrientation};

    #[test]
    fn check_characters_out_of_range() {
        for i in 0..0x10ffff {
            if 0x2800 <= i && i <= 0x28ff { continue; };
            let char = match char::from_u32(i) {
                Some(c) => c,
                // Continue for invalid characters
                None => continue,
            };
            binary_braillie_char_to_index(char, EncodeOrientation::Rows)
              .expect_err("Expected error for invalid char");
        }
    }

    #[test]
    fn check_characters_in_range() {
        for _char in 0x2800..0x28ff {
            let char = char::from_u32(_char).unwrap();
            binary_braillie_char_to_index(char, EncodeOrientation::Rows)
              .expect("Expected no error");
        }
    }
}

#[cfg(test)]
mod test_lib_all {
    use crate::{
        binary_braillie_char_to_index,
        binary_braillie_to_bytes,
        index_to_braillie_char,
        to_binary_braillie,
        EncodeOrientation
    };

    #[test]
    fn check_coding_multiple() {
        for i in 0x00..0xff {
            let mut as_char = index_to_braillie_char(
                i,
                EncodeOrientation::Rows
            );
            for _ in 0..10 {
                let as_index = binary_braillie_char_to_index(
                    as_char,
                    EncodeOrientation::Rows
                ).unwrap();
                assert_eq!(as_index, i);
                as_char = index_to_braillie_char(as_index, EncodeOrientation::Rows)
            }
        }
    }

    #[test]
    fn check_coding_multiple_bytes() {
        for i in 0x00..0xff {
            let mut as_str = index_to_braillie_char(i, EncodeOrientation::Rows).to_string();
            for _ in 0..10 {
                let as_bytes = binary_braillie_to_bytes(
                    as_str,
                    EncodeOrientation::Rows
                ).unwrap();
                assert_eq!(i, as_bytes[0]);
                as_str = to_binary_braillie(
                    &as_bytes,
                    as_bytes.len(),
                    EncodeOrientation::Rows
                ).to_string()
            }
        }
    }
}
